import warnings
from copy import deepcopy
from functools import singledispatch
from typing import Any, Iterable, Type, Union

import numpy as np

from ..parser.openqasm_ast import (
    ArrayLiteral,
    ArrayType,
    BitstringLiteral,
    BitType,
    BooleanLiteral,
    BoolType,
    ClassicalType,
    DiscreteSet,
    Expression,
    FloatLiteral,
    FloatType,
    Identifier,
    IndexedIdentifier,
    IntegerLiteral,
    IntType,
    UintType,
)

LiteralType = Union[BooleanLiteral, IntegerLiteral, FloatLiteral, ArrayLiteral, BitstringLiteral]


@singledispatch
def cast_to(into: Union[ClassicalType, Type[LiteralType]], variable: LiteralType) -> LiteralType:
    """Cast a variable into a given type. Order of parameters is to enable singledispatch"""
    if type(variable) == into:
        return variable
    if into == BooleanLiteral or isinstance(into, BoolType):
        return BooleanLiteral(bool(variable.value))
    if into == IntegerLiteral:
        return IntegerLiteral(int(variable.value))
    if into == FloatLiteral:
        return FloatLiteral(float(variable.value))
    raise TypeError(f"Cannot cast {type(variable).__name__} into {into.__name__}.")


@cast_to.register
def _(
    into: BitType, variable: Union[BooleanLiteral, ArrayLiteral, BitstringLiteral]
) -> ArrayLiteral:
    """
    Bit types can be sized or not, represented as Boolean literals or Array literals.
    Sized bit types can be instantiated with a Bitstring literal or Array literal.
    """
    if not into.size:
        return cast_to(BooleanLiteral, variable)
    size = into.size.value
    if isinstance(variable, BitstringLiteral):
        variable = convert_string_to_bool_array(variable)
    if (
        not all(isinstance(x, BooleanLiteral) for x in variable.values)
        or len(variable.values) != size
    ):
        raise ValueError(f"Invalid array to cast to bit register of size {size}: {variable}.")
    return ArrayLiteral(deepcopy(variable.values))


@cast_to.register
def _(into: IntType, variable: LiteralType) -> IntegerLiteral:
    """Cast to int with overflow warnings"""
    if isinstance(variable, ArrayLiteral):
        value = int("".join("01"[x.value] for x in variable.values[1:]), base=2)
        if variable.values[0].value:
            value *= -1
    else:
        value = variable.value
        if into.size is not None:
            limit = 2 ** (into.size.value - 1)
            value = int(np.sign(value) * (np.abs(int(value)) % limit))
            if value != variable.value:
                warnings.warn(
                    f"Integer overflow for value {variable.value} and size {into.size.value}."
                )
    return IntegerLiteral(value)


@cast_to.register
def _(into: UintType, variable: LiteralType) -> IntegerLiteral:
    """Cast to uint with overflow warnings. Bit registers can be cast to uint."""
    if isinstance(variable, ArrayLiteral):
        return IntegerLiteral(int("".join("01"[x.value] for x in variable.values), base=2))
    value = variable.value
    if into.size is not None:
        limit = 2**into.size.value
        value = int(value) % limit
        if value != variable.value:
            warnings.warn(
                f"Unsigned integer overflow for value {variable.value} and size {into.size.value}."
            )
    return IntegerLiteral(value)


@cast_to.register
def _(into: FloatType, variable: LiteralType) -> FloatLiteral:
    """Cast to float"""
    if into.size is None:
        value = float(variable.value)
    else:
        if into.size.value not in (16, 32, 64):
            raise ValueError("Float size must be one of {16, 32, 64}.")
        value = float(np.array(variable.value, dtype=np.dtype(f"float{into.size.value}")))
    return FloatLiteral(value)


@cast_to.register
def _(into: ArrayType, variable: Union[ArrayLiteral, DiscreteSet]) -> ArrayLiteral:
    """Cast to Array and enforce dimensions"""
    if len(variable.values) != into.dimensions[0].value:
        raise ValueError(
            f"Size mismatch between dimension of size {into.dimensions[0].value} "
            f"and values length {len(variable.values)}"
        )
    subtype = (
        ArrayType(into.base_type, into.dimensions[1:])
        if len(into.dimensions) > 1
        else into.base_type
    )
    return ArrayLiteral([cast_to(subtype, v) for v in variable.values])


def is_literal(expression: Expression) -> bool:
    return isinstance(
        expression,
        (
            BooleanLiteral,
            IntegerLiteral,
            FloatLiteral,
            BitstringLiteral,
            ArrayLiteral,
        ),
    )


def convert_string_to_bool_array(bit_string: BitstringLiteral) -> ArrayLiteral:
    """Convert BitstringLiteral to Boolean ArrayLiteral"""
    return ArrayLiteral(
        [BooleanLiteral(x == "1") for x in np.binary_repr(bit_string.value, bit_string.width)]
    )


def convert_bool_array_to_string(bit_string: ArrayLiteral) -> str:
    """Convert Boolean ArrayLiteral into a binary string"""
    return "".join(("1" if x.value else "0") for x in bit_string.values)


def is_none_like(value: Any) -> bool:
    """Returns whether value is None or an Array of Nones"""
    if isinstance(value, ArrayLiteral):
        return all(is_none_like(v) for v in value.values)
    return value is None


@singledispatch
def get_identifier_name(identifier: Union[Identifier, IndexedIdentifier]) -> str:
    """Get name of an identifier"""
    return identifier.name


@get_identifier_name.register
def _(identifier: IndexedIdentifier) -> str:
    """Get name of an indexed identifier"""
    return identifier.name.name


@singledispatch
def wrap_value_into_literal(value: Any) -> LiteralType:
    """Wrap a primitive variable into an AST node"""
    raise TypeError(f"Cannot wrap {value} into literal type")


@wrap_value_into_literal.register
def _(value: int) -> IntegerLiteral:
    return IntegerLiteral(value)


@wrap_value_into_literal.register
def _(value: float) -> FloatLiteral:
    return FloatLiteral(value)


@wrap_value_into_literal.register
def _(value: bool) -> BooleanLiteral:
    return BooleanLiteral(value)


@wrap_value_into_literal.register
def _(value: str) -> BitstringLiteral:
    return BitstringLiteral(value=int(value, base=2), width=len(value))


@wrap_value_into_literal.register(list)
def _(value: Iterable[Any]) -> ArrayLiteral:
    return ArrayLiteral([wrap_value_into_literal(v) for v in value])



# Generated from qasm3Parser.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .qasm3Parser import qasm3Parser
else:
    from qasm3Parser import qasm3Parser

# This class defines a complete generic visitor for a parse tree produced by qasm3Parser.

class qasm3ParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by qasm3Parser#program.
    def visitProgram(self, ctx:qasm3Parser.ProgramContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#version.
    def visitVersion(self, ctx:qasm3Parser.VersionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#statement.
    def visitStatement(self, ctx:qasm3Parser.StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#annotation.
    def visitAnnotation(self, ctx:qasm3Parser.AnnotationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#scope.
    def visitScope(self, ctx:qasm3Parser.ScopeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#pragma.
    def visitPragma(self, ctx:qasm3Parser.PragmaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#statementOrScope.
    def visitStatementOrScope(self, ctx:qasm3Parser.StatementOrScopeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#calibrationGrammarStatement.
    def visitCalibrationGrammarStatement(self, ctx:qasm3Parser.CalibrationGrammarStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#includeStatement.
    def visitIncludeStatement(self, ctx:qasm3Parser.IncludeStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#breakStatement.
    def visitBreakStatement(self, ctx:qasm3Parser.BreakStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#continueStatement.
    def visitContinueStatement(self, ctx:qasm3Parser.ContinueStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#endStatement.
    def visitEndStatement(self, ctx:qasm3Parser.EndStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#forStatement.
    def visitForStatement(self, ctx:qasm3Parser.ForStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#ifStatement.
    def visitIfStatement(self, ctx:qasm3Parser.IfStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#returnStatement.
    def visitReturnStatement(self, ctx:qasm3Parser.ReturnStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#whileStatement.
    def visitWhileStatement(self, ctx:qasm3Parser.WhileStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#barrierStatement.
    def visitBarrierStatement(self, ctx:qasm3Parser.BarrierStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#boxStatement.
    def visitBoxStatement(self, ctx:qasm3Parser.BoxStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#delayStatement.
    def visitDelayStatement(self, ctx:qasm3Parser.DelayStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#gateCallStatement.
    def visitGateCallStatement(self, ctx:qasm3Parser.GateCallStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#measureArrowAssignmentStatement.
    def visitMeasureArrowAssignmentStatement(self, ctx:qasm3Parser.MeasureArrowAssignmentStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#resetStatement.
    def visitResetStatement(self, ctx:qasm3Parser.ResetStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#aliasDeclarationStatement.
    def visitAliasDeclarationStatement(self, ctx:qasm3Parser.AliasDeclarationStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#classicalDeclarationStatement.
    def visitClassicalDeclarationStatement(self, ctx:qasm3Parser.ClassicalDeclarationStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#constDeclarationStatement.
    def visitConstDeclarationStatement(self, ctx:qasm3Parser.ConstDeclarationStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#ioDeclarationStatement.
    def visitIoDeclarationStatement(self, ctx:qasm3Parser.IoDeclarationStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#oldStyleDeclarationStatement.
    def visitOldStyleDeclarationStatement(self, ctx:qasm3Parser.OldStyleDeclarationStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#quantumDeclarationStatement.
    def visitQuantumDeclarationStatement(self, ctx:qasm3Parser.QuantumDeclarationStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#defStatement.
    def visitDefStatement(self, ctx:qasm3Parser.DefStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#externStatement.
    def visitExternStatement(self, ctx:qasm3Parser.ExternStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#gateStatement.
    def visitGateStatement(self, ctx:qasm3Parser.GateStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#assignmentStatement.
    def visitAssignmentStatement(self, ctx:qasm3Parser.AssignmentStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#expressionStatement.
    def visitExpressionStatement(self, ctx:qasm3Parser.ExpressionStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#defcalStatement.
    def visitDefcalStatement(self, ctx:qasm3Parser.DefcalStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#bitwiseXorExpression.
    def visitBitwiseXorExpression(self, ctx:qasm3Parser.BitwiseXorExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#additiveExpression.
    def visitAdditiveExpression(self, ctx:qasm3Parser.AdditiveExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#durationofExpression.
    def visitDurationofExpression(self, ctx:qasm3Parser.DurationofExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#parenthesisExpression.
    def visitParenthesisExpression(self, ctx:qasm3Parser.ParenthesisExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#comparisonExpression.
    def visitComparisonExpression(self, ctx:qasm3Parser.ComparisonExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#multiplicativeExpression.
    def visitMultiplicativeExpression(self, ctx:qasm3Parser.MultiplicativeExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#logicalOrExpression.
    def visitLogicalOrExpression(self, ctx:qasm3Parser.LogicalOrExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#castExpression.
    def visitCastExpression(self, ctx:qasm3Parser.CastExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#powerExpression.
    def visitPowerExpression(self, ctx:qasm3Parser.PowerExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#bitwiseOrExpression.
    def visitBitwiseOrExpression(self, ctx:qasm3Parser.BitwiseOrExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#callExpression.
    def visitCallExpression(self, ctx:qasm3Parser.CallExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#bitshiftExpression.
    def visitBitshiftExpression(self, ctx:qasm3Parser.BitshiftExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#bitwiseAndExpression.
    def visitBitwiseAndExpression(self, ctx:qasm3Parser.BitwiseAndExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#equalityExpression.
    def visitEqualityExpression(self, ctx:qasm3Parser.EqualityExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#logicalAndExpression.
    def visitLogicalAndExpression(self, ctx:qasm3Parser.LogicalAndExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#indexExpression.
    def visitIndexExpression(self, ctx:qasm3Parser.IndexExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#unaryExpression.
    def visitUnaryExpression(self, ctx:qasm3Parser.UnaryExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#literalExpression.
    def visitLiteralExpression(self, ctx:qasm3Parser.LiteralExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#aliasExpression.
    def visitAliasExpression(self, ctx:qasm3Parser.AliasExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#declarationExpression.
    def visitDeclarationExpression(self, ctx:qasm3Parser.DeclarationExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#measureExpression.
    def visitMeasureExpression(self, ctx:qasm3Parser.MeasureExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#rangeExpression.
    def visitRangeExpression(self, ctx:qasm3Parser.RangeExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#setExpression.
    def visitSetExpression(self, ctx:qasm3Parser.SetExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#arrayLiteral.
    def visitArrayLiteral(self, ctx:qasm3Parser.ArrayLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#indexOperator.
    def visitIndexOperator(self, ctx:qasm3Parser.IndexOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#indexedIdentifier.
    def visitIndexedIdentifier(self, ctx:qasm3Parser.IndexedIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#returnSignature.
    def visitReturnSignature(self, ctx:qasm3Parser.ReturnSignatureContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#gateModifier.
    def visitGateModifier(self, ctx:qasm3Parser.GateModifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#scalarType.
    def visitScalarType(self, ctx:qasm3Parser.ScalarTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#qubitType.
    def visitQubitType(self, ctx:qasm3Parser.QubitTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#arrayType.
    def visitArrayType(self, ctx:qasm3Parser.ArrayTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#arrayReferenceType.
    def visitArrayReferenceType(self, ctx:qasm3Parser.ArrayReferenceTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#designator.
    def visitDesignator(self, ctx:qasm3Parser.DesignatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#gateOperand.
    def visitGateOperand(self, ctx:qasm3Parser.GateOperandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#externArgument.
    def visitExternArgument(self, ctx:qasm3Parser.ExternArgumentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#defcalArgument.
    def visitDefcalArgument(self, ctx:qasm3Parser.DefcalArgumentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#argumentDefinition.
    def visitArgumentDefinition(self, ctx:qasm3Parser.ArgumentDefinitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#argumentDefinitionList.
    def visitArgumentDefinitionList(self, ctx:qasm3Parser.ArgumentDefinitionListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#expressionList.
    def visitExpressionList(self, ctx:qasm3Parser.ExpressionListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#defcalArgumentList.
    def visitDefcalArgumentList(self, ctx:qasm3Parser.DefcalArgumentListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#identifierList.
    def visitIdentifierList(self, ctx:qasm3Parser.IdentifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#gateOperandList.
    def visitGateOperandList(self, ctx:qasm3Parser.GateOperandListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by qasm3Parser#externArgumentList.
    def visitExternArgumentList(self, ctx:qasm3Parser.ExternArgumentListContext):
        return self.visitChildren(ctx)



del qasm3Parser


# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

import numpy as np


class Operation(ABC):
    """
    Encapsulates an operation acting on a set of target qubits.
    """

    @property
    @abstractmethod
    def targets(self) -> Tuple[int, ...]:
        """Tuple[int, ...]: The indices of the qubits the operation applies to.

        Note: For an index to be a target of an observable, the observable must have a nontrivial
        (i.e. non-identity) action on that index. For example, a tensor product observable with a
        Z factor on qubit j acts trivially on j, so j would not be a target. This does not apply to
        gate operations.
        """


class GateOperation(Operation, ABC):
    """
    Encapsulates a unitary quantum gate operation acting on
    a set of target qubits.
    """

    @property
    @abstractmethod
    def matrix(self) -> np.ndarray:
        """np.ndarray: The matrix representation of the operation."""

    def __eq__(self, other):
        return self.targets == other.targets and np.allclose(self.matrix, other.matrix)


class KrausOperation(Operation, ABC):
    """
    Encapsulates a quantum channel acting on a set of target qubits in the Kraus operator
    representation.
    """

    @property
    @abstractmethod
    def matrices(self) -> List[np.ndarray]:
        """List[np.ndarray]: A list of matrices representing Kraus operators."""

    def __eq__(self, other):
        return self.targets == other.targets and np.allclose(self.matrices, other.matrices)


class Observable(Operation, ABC):
    """
    Encapsulates an observable to be measured in the computational basis.
    """

    @property
    def measured_qubits(self) -> Tuple[int, ...]:
        """Tuple[int, ...]: The indices of the qubits that are measured for this observable.

        Unlike `targets`, this includes indices on which the observable acts trivially.
        For example, a tensor product observable made entirely of n Z factors will have
        n measured qubits.
        """
        return self.targets

    @property
    def is_standard(self) -> bool:
        r"""bool: Whether the observable is Pauli-like, that is, has eigenvalues of :math:`\pm 1`.

        Examples include the Pauli and Hadamard observables.
        """
        return False

    def __pow__(self, power: int) -> Observable:
        if not isinstance(power, int):
            raise TypeError("power must be integer")
        return self._pow(power)

    @abstractmethod
    def _pow(self, power: int) -> Observable:
        """Raises this observable to the given power.

        Only defined for integer powers.

        Args:
            power (int): The power to raise the observable to.

        Returns:
            Observable: The observable raised to the given power.
        """

    @property
    @abstractmethod
    def eigenvalues(self) -> np.ndarray:
        """
        np.ndarray: The eigenvalues of the observable ordered by computational basis state.
        """

    @abstractmethod
    def apply(self, state: np.ndarray) -> np.ndarray:
        """Applies this observable to the given state.

        Args:
            state (np.ndarray): The state to apply the observable to.

        Returns:
            np.ndarray: The state after the observable has been applied.
        """

    @abstractmethod
    def fix_qubit(self, qubit: int) -> Observable:
        """Creates a copy of it acting on the given qubit.

        Only defined for observables that act on 1 qubit.

        Args:
            qubit (int): The target qubit of the new observable.

        Returns:
            Observable: A copy of this observable, acting on the new qubit.
        """

    @abstractmethod
    def diagonalizing_gates(self, num_qubits: Optional[int] = None) -> Tuple[GateOperation, ...]:
        """The gates that diagonalize the observable in the computational basis.

        Args:
            num_qubits (int, optional): The number of qubits the observable acts on.
                Only used if no target is specified, in which case a gate is created
                for each target qubit. This only makes sense for single-qubit observables.

        Returns:
            Tuple[GateOperation, ...]: The gates that diagonalize the observable in the
            computational basis, if it is not already in the computational basis.
            If there is no explicit target, this method returns a tuple of gates
            acting on every qubit.
        """



from functools import singledispatch, update_wrapper


def singledispatchmethod(func):
    """Implement singledispatchmethod for Python 3.7"""
    dispatcher = singledispatch(func)

    def wrapper(*args, **kw):
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)

    wrapper.register = dispatcher.register
    update_wrapper(wrapper, func)
    return wrapper
